/* $Id$ */

/*
 * Copyright (c) 2006
 *               Eino Tuominen <eino@utu.fi>
 *               Antti Siira <antti@utu.fi>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "common.h"
#include "msgqueue.h"

int client_sjsms(int argc, char **argv);
int client_postfix(int argc, char **argv);

int
main(int argc, char **argv)
{
	if (argc < 2) {
		fprintf(stderr, "usage gclient PROTOCOL\n");
		return 1;
	}

	if (strcmp(argv[1], "postfix") == 0)
		return client_postfix(argc, argv);
	else if (strcmp(argv[1], "sjsms") == 0)
		return client_sjsms(argc, argv);
	else {
		fprintf(stderr, "supported protocols: 'sjsms' and 'postfix'\n");
		return 1;
	}
}
